@@ -1,3 +1,4 @@
+# -*- coding: utf-8 -*-
 # docs/COPYING 2a + DRY: https://github.com/getmail6/getmail6
 # Please refer to the git history regarding who changed what and when in this file.
 
@@ -5,6 +6,25 @@
 
 '''
 
+import sys
+import os
+import time
+import signal
+import types
+import codecs
+from collections import namedtuple
+import tempfile
+import errno
+from threading import Condition
+import socket
+
+from argparse import Namespace
+import subprocess
+
+from getmailcore.exceptions import *
+import getmailcore.logging
+from getmailcore.utilities import *
+
 __all__ = [
     'run_command',
     'ConfigurableBase',
@@ -23,24 +43,6 @@ __all__ = [
     'ConfMboxPath',
 ]
 
-import sys
-import os
-import time
-import signal
-import types
-import codecs
-from collections import namedtuple
-import tempfile
-import errno
-from threading import Condition
-
-from argparse import Namespace
-import subprocess
-
-from getmailcore.exceptions import *
-import getmailcore.logging
-from getmailcore.utilities import *
-
 if sys.version_info.major > 2:
     TemporaryFile23 = lambda: tempfile.TemporaryFile('bw+')
 else:
@@ -299,7 +301,10 @@ class ConfMboxPath(ConfString):
             )
         fd = os.open(val, os.O_RDWR)
         status_old = os.fstat(fd)
-        f = os.fdopen(fd, 'br+')
+        try:
+            f = os.fdopen(fd, 'br+')
+        except ValueError: # Py2
+            f = os.fdopen(fd, 'r+')
         # Check if it _is_ an mbox file.  mbox files must start with "From "
         # in their first line, or are 0-length files.
         f.seek(0, 0)
@@ -331,7 +336,7 @@ class ConfigurableBase(object):
                      - name - parameter name
                      - type - a type function to compare the parameter value
                        against (i.e. str, int, bool)
-                     - default - optional default value.  If not preseent, the
+                     - default - optional default value.  If not present, the
                        parameter is required.
     '''
 
@@ -339,9 +344,8 @@ class ConfigurableBase(object):
         self.log = getmailcore.logging.Logger()
         self.log.trace()
         self.conf = {}
-        allowed_params = set([item.name for item in self._confitems])
         for (name, value) in args.items():
-            if not name in allowed_params:
+            if name not in self:
                 self.log.warning('Warning: ignoring unknown parameter "%s" '
                                  '(value: %s)\n' % (name, value))
                 continue
@@ -385,6 +389,9 @@ class ConfigurableBase(object):
                 confstring += '%s="%s"' % (name, self.conf[name])
         return confstring
 
+    def __contains__(self, confitem):
+        return confitem in {item.name for item in self._confitems}
+
 #######################################
 
 class ForkingBase(object):
@@ -423,7 +430,7 @@ class ForkingBase(object):
 
     def _wait_for_child(self, childpid):
         self.__child_exited.acquire()
-        if not self.__child_exited.wait(60):
+        if self.__child_exited.wait(socket.getdefaulttimeout() or 60) == False: # Py2, <Py3.2: always None
             raise getmailOperationError('waiting child pid %d timed out'
                                         % childpid)
         self.__child_exited.release()
